@@ -1,7 +1,6 @@
 import tempfile
 from typing import TYPE_CHECKING, Any, Dict, List, Type, Union
 
-import numpy
 import numpy as np
 import pandas as pd
 from numpy.typing import ArrayLike
@@ -22,6 +21,8 @@ from policyengine_core.tracers import (
     SimpleTracer,
     TracingParameterNodeAtInstant,
 )
+import random
+from policyengine_core.tools.hugging_face import *
 
 import json
 
@@ -34,7 +35,9 @@ from policyengine_core.tracers import SimpleTracer
 from policyengine_core.variables import Variable, QuantityType
 from policyengine_core.reforms.reform import Reform
 from policyengine_core.parameters import get_parameter
-from policyengine_core.simulations.sim_macro_cache import SimulationMacroCache
+from policyengine_core.simulations.simulation_macro_cache import (
+    SimulationMacroCache,
+)
 
 
 class Simulation:
@@ -51,7 +54,7 @@ class Simulation:
     default_dataset: Dataset = None
     """The default dataset class to use if none is provided."""
 
-    default_role: str = None
+    default_role: str = "member"
     """The default role to assign people to groups if none is provided."""
 
     default_input_period: str = None
@@ -69,10 +72,10 @@ class Simulation:
     is_over_dataset: bool = False
     """Whether this simulation is built over a dataset."""
 
-    macro_cache_read: bool = True
+    macro_cache_read: bool = False
     """Whether to read from the macro cache."""
 
-    macro_cache_write: bool = True
+    macro_cache_write: bool = False
     """Whether to write to the macro cache."""
 
     start_instant: str = None
@@ -86,7 +89,15 @@ class Simulation:
         dataset: Union[str, Type[Dataset]] = None,
         reform: Reform = None,
         trace: bool = False,
+        default_input_period: str = None,
+        default_calculation_period: str = None,
     ):
+        self.default_input_period = (
+            default_input_period or self.default_input_period
+        )
+        self.default_calculation_period = (
+            default_calculation_period or self.default_calculation_period
+        )
         if tax_benefit_system is None:
             if (
                 self.default_tax_benefit_system_instance is not None
@@ -102,6 +113,7 @@ class Simulation:
         self.reform = reform
         self.tax_benefit_system = tax_benefit_system
         self.branch_name = "default"
+        self.dataset = dataset
 
         if dataset is None:
             if self.default_dataset is not None:
@@ -147,6 +159,18 @@ class Simulation:
 
         if dataset is not None:
             if isinstance(dataset, str):
+                if "hf://" in dataset:
+                    owner, repo, filename = dataset.split("/")[-3:]
+                    if "@" in filename:
+                        version = filename.split("@")[-1]
+                        filename = filename.split("@")[0]
+                    else:
+                        version = None
+                    dataset = download_huggingface_dataset(
+                        repo=f"{owner}/{repo}",
+                        repo_filename=filename,
+                        version=version,
+                    )
                 datasets_by_name = {
                     dataset.name: dataset for dataset in self.datasets
                 }
@@ -208,6 +232,8 @@ class Simulation:
             for subreform in reform:
                 self.apply_reform(subreform)
         else:
+            if isinstance(reform, dict):
+                reform = Reform.from_dict(reform)
             reform.apply(self.tax_benefit_system)
 
     def build_from_populations(
@@ -607,7 +633,7 @@ class Simulation:
 
         smc = SimulationMacroCache(self.tax_benefit_system)
 
-        # Check if cache could be used, if available, check if path exists
+        # Check if cache can be used, if available, check if path exists
         is_cache_available = self.check_macro_cache(variable_name, str(period))
         if is_cache_available:
             smc.set_cache_path(
@@ -631,15 +657,20 @@ class Simulation:
                     return value
 
         if variable.requires_computation_after is not None:
-            if variable.requires_computation_after not in [
+            variables_in_stack = [
                 node.get("name") for node in self.tracer.stack
-            ]:
+            ]
+            variable_in_stack = (
+                variable.requires_computation_after in variables_in_stack
+            )
+            required_is_known_periods = self.get_holder(
+                variable.requires_computation_after
+            ).get_known_periods()
+            if (not variable_in_stack) and (
+                not len(required_is_known_periods) > 0
+            ):
                 raise ValueError(
-                    f"Variable {variable_name} requires {variable.requires_computation_after} to be requested first. The full stack is: "
-                    + "\n".join(
-                        f"  - {node.get('name')} {node.get('period')}, {node.get('branch_name')}"
-                        for node in self.tracer.stack
-                    )
+                    f"Variable {variable_name} requires {variable.requires_computation_after} to be requested first. That variable is known in: {required_is_known_periods}. The full stack is: {variables_in_stack}. {variable_in_stack, len(required_is_known_periods) > 0}"
                 )
         alternate_period_handling = False
         if variable.definition_period == MONTH and period.unit == YEAR:
@@ -735,7 +766,15 @@ class Simulation:
                     array = holder.default_array()
 
             if variable.defined_for is not None:
-                array = np.where(mask, array, np.zeros_like(array))
+                array = np.where(mask, array, variable.default_value)
+                if variable.value_type == Enum:
+                    array = np.array(
+                        [
+                            item.index if isinstance(item, Enum) else item
+                            for item in array
+                        ]
+                    )
+                    array = EnumArray(array, variable.possible_values)
 
             array = self._cast_formula_result(array, variable)
             holder.put_in_cache(array, period, self.branch_name)
@@ -1013,7 +1052,7 @@ class Simulation:
         if variable.value_type == Enum and not isinstance(value, EnumArray):
             return variable.possible_values.encode(value)
 
-        if not isinstance(value, numpy.ndarray):
+        if not isinstance(value, np.ndarray):
             population = self.get_variable_population(variable.name)
             value = population.filled_array(value)
 
@@ -1418,14 +1457,23 @@ class Simulation:
         """
         Check if the variable is able to have cached value
         """
-        if (
-            hasattr(self, "dataset")
-            and self.dataset.data_format == Dataset.FLAT_FILE
-        ):
+        if not self.macro_cache_read:
             return False
 
-        if self.is_over_dataset:
-            return True
+        # Dataset should always exist, but just in case
+        if not hasattr(self, "dataset"):
+            return False
+
+        # If no dataset, no need to cache
+        if self.dataset is None:
+            return False
+
+        # If using a flat file dataset, we're unable to cache
+        if self.dataset.data_format == Dataset.FLAT_FILE:
+            return False
+
+        if not self.is_over_dataset:
+            return False
 
         variable = self.tax_benefit_system.get_variable(variable_name)
         parameter_deps = variable.exhaustive_parameter_dependencies
@@ -1454,13 +1502,160 @@ class Simulation:
         df = pd.DataFrame()
 
         for variable in self.tax_benefit_system.variables:
+            variable_meta = self.tax_benefit_system.variables[variable]
             for period in self.get_holder(variable).get_known_periods():
+                # Test if period matches entity definition period
+                if variable_meta.definition_period != period.unit:
+                    continue
                 values = self.calculate(variable, period, map_to="person")
                 if values is not None:
                     df[f"{variable}__{period}"] = values
 
         return df
 
+    def to_input_dict(self) -> dict:
+        """Exports a dictionary which can be loaded back to a new Simulation to reproduce the same results.
+
+        Returns:
+            dict: The dictionary containing the input values.
+        """
+        data = {}
+
+        for variable in self.tax_benefit_system.variables:
+            data[variable] = {}
+            for period in self.get_holder(variable).get_known_periods():
+                values = self.calculate(variable, period, map_to="person")
+                if values is not None:
+                    data[variable][str(period)] = values.tolist()
+
+            if len(data[variable]) == 0:
+                del data[variable]
+
+        return data
+
+    def subsample(
+        self,
+        n=None,
+        frac=None,
+        seed=None,
+        time_period=None,
+        quantize_weights: bool = True,
+    ) -> "Simulation":
+        """Quantize the simulation to a smaller size by sampling households.
+
+        Args:
+            n (int, optional): The number of households to sample. Defaults to 10_000.
+            frac (float, optional): The fraction of households to sample. Defaults to None.
+            seed (int, optional): The key used to seed the random number generator. Defaults to the dataset name.
+            time_period (str, optional): Sample households based on their weight in this time period. Defaults to the default calculation period.
+
+        Returns:
+            Simulation: The quantized simulation.
+        """
+        default_calculation_period = self.default_calculation_period
+        # Set default key if not provided
+        if seed is None:
+            seed = self.dataset.name
+
+        # Set default time period if not provided
+        if time_period is None:
+            time_period = self.default_calculation_period
+
+        # Convert simulation inputs to DataFrame
+        df = self.to_input_dataframe()
+
+        # Extract time period from DataFrame columns
+        df_time_period = (
+            df.columns[df.columns.str.contains("household_id__")]
+            .values[0]
+            .split("__")[1]
+        )
+        df_household_id_column = f"household_id__{df_time_period}"
+        df_person_id_column = f"person_id__{df_time_period}"
+
+        # Determine the appropriate household weight column
+        if f"household_weight__{time_period}" in df.columns:
+            household_weight_column = f"household_weight__{time_period}"
+        else:
+            household_weight_column = f"household_weight__{df_time_period}"
+
+        # Group by household ID and get the first entry for each group
+        h_df = df.groupby(df_household_id_column).first()
+        h_ids = pd.Series(h_df.index)
+        if n is None and frac is None:
+            raise ValueError("Either n or frac must be provided.")
+        if n is None:
+            n = int(len(h_ids) * frac)
+        h_weights = pd.Series(h_df[household_weight_column].values)
+
+        frac = n / len(h_ids)
+
+        # Seed the random number generators for reproducibility
+        random.seed(str(seed))
+        state = random.randint(0, 2**32 - 1)
+        np.random.seed(state)
+
+        h_ids = h_ids[h_weights > 0]
+        h_weights = h_weights[h_weights > 0]
+
+        # Sample household IDs based on their weights
+        chosen_household_ids = pd.Series(
+            np.random.choice(
+                h_ids,
+                n,
+                p=(
+                    h_weights.values / h_weights.values.sum()
+                    if quantize_weights
+                    else None
+                ),
+                replace=True,
+            )
+        )
+
+        household_id_to_count = {}
+        for household_id in chosen_household_ids:
+            if household_id not in household_id_to_count:
+                household_id_to_count[household_id] = 0
+            household_id_to_count[household_id] += 1
+
+        subset_df = df[
+            df[df_household_id_column].isin(chosen_household_ids)
+        ].copy()
+
+        household_counts = subset_df[df_household_id_column].map(
+            lambda x: household_id_to_count.get(x, 0)
+        )
+
+        # Adjust household weights to maintain the total weight
+
+        for col in subset_df.columns:
+            if "weight__" in col:
+                target_total_weight = df[col].values.sum()
+                if not quantize_weights:
+                    subset_df[col] *= household_counts.values
+                else:
+                    subset_df[col] = household_counts.values
+                subset_df[col] *= (
+                    target_total_weight / subset_df[col].values.sum()
+                )
+
+        df = subset_df
+
+        # Update the dataset and rebuild the simulation
+        self.dataset = Dataset.from_dataframe(df, self.dataset.time_period)
+        self.build_from_dataset()
+
+        # Ensure the baseline branch has the new data.
+        if "baseline" in self.branches:
+            baseline_tax_benefit_system = self.branches[
+                "baseline"
+            ].tax_benefit_system
+            self.branches["baseline"] = self.clone()
+            self.branches["tax_benefit_system"] = baseline_tax_benefit_system
+
+        self.default_calculation_period = default_calculation_period
+        return self
+
 
 class NpEncoder(json.JSONEncoder):
     def default(self, obj):
